@@ -16,6 +16,7 @@ import re
 import shutil
 import sys
 import time
+import traceback
 from abc import ABC, abstractmethod
 
 try:
@@ -68,6 +69,16 @@ PARTITION_STRATEGY = {
 }
 
 
+class SchemaResolverVersion(Enum):
+    """The schema resolver version to use."""
+
+    V1 = 1
+    V2 = 2
+
+    def __str__(self) -> str:
+        return str(self.value)
+
+
 @dataclass
 class BigQueryTable:
     name: str
@@ -82,6 +93,7 @@ class BigQueryTable:
     """The ingestion strategy for this table."""
     transforms: Dict[str, bool] = field(default_factory=dict)
     """A dict of transformation rules to apply to the table schema."""
+    schema_resolver_version: SchemaResolverVersion = SchemaResolverVersion.V1
 
     @property
     def schema_translator(self) -> "SchemaTranslator":
@@ -90,6 +102,7 @@ class BigQueryTable:
             self._schema_translator = SchemaTranslator(
                 schema=self.jsonschema,
                 transforms=self.transforms,
+                resolver_version=self.schema_resolver_version,
             )
         return self._schema_translator
 
@@ -99,6 +112,10 @@ class BigQueryTable:
             return self.schema_translator.translated_schema_transformed
         return self.schema_translator.translated_schema
 
+    def get_escaped_name(self, suffix: str = '') -> str:
+        """Returns the table name as as escaped SQL string."""
+        return f"`{self.project}`.`{self.dataset}`.`{self.name}{suffix}`"
+
     def get_resolved_schema(self, apply_transforms: bool = False) -> List[bigquery.SchemaField]:
         """Returns the schema for this table after factoring in the ingestion strategy."""
         if self.ingestion_strategy is IngestionStrategy.FIXED:
@@ -185,9 +202,9 @@ class BigQueryTable:
             "description": (
                 "Generated by target-bigquery.\nStream Schema\n{schema}\nBigQuery Ingestion"
                 " Strategy: {strategy}".format(
-                    schema=(schema_dump[:16000] + "...")
-                    if len(schema_dump) > 16000
-                    else schema_dump,
+                    schema=(
+                        (schema_dump[:16000] + "...") if len(schema_dump) > 16000 else schema_dump
+                    ),
                     strategy=self.ingestion_strategy,
                 )
             ),
@@ -225,18 +242,31 @@ class BaseWorker(ABC):
         ext_id: str,
         queue: "Queue",
         credentials: BigQueryCredentials,
+        # 3 connections for job, error, and non-error log notifications
         job_notifier: "Connection",
+        error_notifier: "Connection",
+        log_notifier: "Connection",
     ):
+        """Initialize a worker."""
         super().__init__()
         self.ext_id: str = ext_id
         self.queue: "Queue" = queue
         self.credentials: BigQueryCredentials = credentials
         self.job_notifier: "Connection" = job_notifier
+        self.error_notifier: "Connection" = error_notifier
+        self.log_notifier: "Connection" = log_notifier
 
     @abstractmethod
     def run(self) -> None:
+        """Run the worker."""
         raise NotImplementedError
 
+    def serialize_exception(self, exc: Exception) -> str:
+        """Serialize an exception to a string."""
+        msg = "".join(traceback.format_exception(type(exc), exc, exc.__traceback__, chain=False))
+        msg += f"\nWorker ID: {self.ext_id}\n"
+        return msg
+
 
 # Base class for sinks which use a fixed schema, optionally
 # including a view which is created on top of the table to unpack the data
@@ -267,6 +297,7 @@ class BaseBigQuerySink(BatchSink):
             "jsonschema": self.schema,
             "transforms": self.config.get("column_name_transforms", {}),
             "ingestion_strategy": self.ingestion_strategy,
+            "schema_resolver_version": SchemaResolverVersion(self.config.get("schema_resolver_version", 1)),
         }
         self.table = BigQueryTable(name=self.table_name, **opts)
         self.create_target(key_properties=key_properties)
@@ -290,7 +321,11 @@ class BaseBigQuerySink(BatchSink):
                     "table": {
                         "expires": datetime.datetime.now() + datetime.timedelta(days=1),
                     },
-                    "dataset": {},
+                    "dataset": {
+                        "location": self.config.get(
+                            "location", BigQueryTable.default_dataset_options()["location"]
+                        )
+                    },
                 },
             )
             time.sleep(2.5)  # Wait for eventual consistency
@@ -304,7 +339,11 @@ class BaseBigQuerySink(BatchSink):
                     "table": {
                         "expires": datetime.datetime.now() + datetime.timedelta(days=1),
                     },
-                    "dataset": {},
+                    "dataset": {
+                        "location": self.config.get(
+                            "location", BigQueryTable.default_dataset_options()["location"]
+                        )
+                    },
                 },
             )
             time.sleep(2.5)  # Wait for eventual consistency
@@ -346,8 +385,7 @@ class BaseBigQuerySink(BatchSink):
         return overwrite_candidate
 
     def _is_dedupe_before_upsert_candidate(self) -> bool:
-        """Determine if this stream is a dedupe before upsert candidate based on user configuration.
-        """
+        """Determine if this stream is a dedupe before upsert candidate based on user configuration."""
         # TODO: we can enable this for `overwrite` too if we want but the purpose would
         # be less for functional reasons (merge constraints) and more for convenience
         dedupe_before_upsert_selection = self.config.get("dedupe_before_upsert", False)
@@ -432,7 +470,7 @@ class BaseBigQuerySink(BatchSink):
         if self.generate_view:
             self.client.query(
                 self.table.schema_translator.generate_view_statement(
-                    str(self.table),
+                    self.table,
                 )
             ).result()
 
@@ -462,13 +500,13 @@ class BaseBigQuerySink(BatchSink):
             if self._is_dedupe_before_upsert_candidate():
                 # We can't use MERGE with a non-unique key, so we need to dedupe the temp table into
                 # a _SESSION scoped intermediate table.
-                tmp = f"{target.name}__tmp"
+                tmp = f"{self.merge_target.name}__tmp"
                 dedupe_query = (
-                    f"SELECT * FROM {self.table} "
+                    f"SELECT * FROM {self.table.get_escaped_name()} "
                     f"QUALIFY ROW_NUMBER() OVER (PARTITION BY {', '.join(self.key_properties)} "
                     f"ORDER BY COALESCE({', '.join(date_columns)}) DESC) = 1"
                 )
-                ctas_tmp = f"CREATE OR REPLACE TEMP TABLE {tmp} AS {dedupe_query}"
+                ctas_tmp = f"CREATE OR REPLACE TEMP TABLE `{tmp}` AS {dedupe_query}"
             merge_clause = (
                 f"MERGE `{self.merge_target}` AS target USING `{tmp or self.table}` AS source ON "
                 + " AND ".join(f"target.`{f}` = source.`{f}`" for f in self.key_properties)
@@ -477,14 +515,14 @@ class BaseBigQuerySink(BatchSink):
                 f"target.`{f.name}` = source.`{f.name}`" for f in target.schema
             )
             insert_clause = (
-                f"INSERT ({', '.join(f.name for f in target.schema)}) "
+                f"INSERT ({', '.join(f'`{f.name}`' for f in target.schema)}) "
                 f"VALUES ({', '.join(f'source.`{f.name}`' for f in target.schema)})"
             )
             self.client.query(
                 f"{ctas_tmp}; {merge_clause} "
                 f"WHEN MATCHED THEN {update_clause} "
                 f"WHEN NOT MATCHED THEN {insert_clause}; "
-                f"DROP TABLE IF EXISTS {self.table};"
+                f"DROP TABLE IF EXISTS {self.table.get_escaped_name()};"
             ).result()
             self.table = self.merge_target
             self.merge_target = None
@@ -493,10 +531,10 @@ class BaseBigQuerySink(BatchSink):
             # Do it in a transaction to avoid partial writes.
             target = self.overwrite_target.as_table()
             self.client.query(
-                f"DROP TABLE IF EXISTS {self.overwrite_target}; "
-                f"CREATE TABLE {self.overwrite_target} LIKE {self.table} AS "
-                f"SELECT * FROM {self.table};"
-                f"DROP TABLE IF EXISTS {self.table};"
+                f"DROP TABLE IF EXISTS {self.overwrite_target.get_escaped_name()}; CREATE TABLE"
+                f" {self.overwrite_target.get_escaped_name()} AS SELECT * FROM"
+                f" {self.table.get_escaped_name()}; DROP TABLE IF EXISTS"
+                f" {self.table.get_escaped_name()};"
             ).result()
             self.table = self.merge_target
             self.merge_target = None
@@ -553,9 +591,13 @@ def augmented_syspath(new_paths: Optional[Iterable[str]] = None):
 def bigquery_client_factory(creds: BigQueryCredentials) -> bigquery.Client:
     """Get a BigQuery client."""
     if creds.path:
-        return bigquery.Client.from_service_account_json(creds.path)
+        return bigquery.Client.from_service_account_json(creds.path,
+            project=creds.project
+        )
     elif creds.json:
-        return bigquery.Client.from_service_account_info(json.loads(creds.json))
+        return bigquery.Client.from_service_account_info(
+            json.loads(creds.json), project=creds.project
+        )
     return bigquery.Client(project=creds.project)
 
 
@@ -563,9 +605,11 @@ def bigquery_client_factory(creds: BigQueryCredentials) -> bigquery.Client:
 def gcs_client_factory(creds: BigQueryCredentials) -> storage.Client:
     """Get a GCS client."""
     if creds.path:
-        return storage.Client.from_service_account_json(creds.path)
+        return storage.Client.from_service_account_json(creds.path, project=creds.project)
     elif creds.json:
-        return storage.Client.from_service_account_info(json.loads(creds.json))
+        return storage.Client.from_service_account_info(
+            json.loads(creds.json), project=creds.project
+        )
     return storage.Client(project=creds.project)
 
 
@@ -582,15 +626,30 @@ def storage_client_factory(
         )
     return bigquery_storage_v1.BigQueryWriteClient()
 
+@dataclass
+class _FieldProjection:
+    projection: str
+    alias: str
+
+    def as_sql(self) -> str:
+        """Return the SQL representation of this projection"""
+        return f'{self.projection} as {self.alias.lstrip()},\n'
+
 
 # This class translates a JSON schema into a BigQuery schema.
 # It also uses the translated schema to generate a CREATE VIEW statement.
 class SchemaTranslator:
     """Translate a JSON schema into a BigQuery schema."""
 
-    def __init__(self, schema: Dict[str, Any], transforms: Dict[str, bool]):
+    def __init__(
+        self,
+        schema: Dict[str, Any],
+        transforms: Dict[str, bool],
+        resolver_version: SchemaResolverVersion = SchemaResolverVersion.V1,
+    ) -> None:
         self.schema = schema
         self.transforms = transforms
+        self.resolver_version = resolver_version
         # Used by fixed schema strategy where we defer transformation
         # to the view statement
         self._translated_schema = None
@@ -639,46 +698,90 @@ class SchemaTranslator:
                 output[k] = self.translate_record(v)
         return output
 
-    def generate_view_statement(self, table_name: str) -> str:
+    def generate_view_statement(self, table_name: BigQueryTable) -> str:
         """Generate a CREATE VIEW statement for the SchemaTranslator `schema`."""
         projection = ""
         for field_ in self.translated_schema[:]:
             if field_.mode == "REPEATED":
                 projection += indent(self._wrap_json_array(field_, path="$", depth=1), " " * 4)
             else:
-                projection += indent(self._bigquery_field_to_projection(field_), " " * 4)
+                projection += indent(self._bigquery_field_to_projection(field_).as_sql(), " " * 4)
 
         return (
-            f"CREATE OR REPLACE VIEW {table_name}_view AS \nSELECT \n{projection} FROM {table_name}"
+            f"CREATE OR REPLACE VIEW {table_name.get_escaped_name('_view')} AS \nSELECT \n{projection} FROM {table_name.get_escaped_name()}"
         )
 
     def _jsonschema_property_to_bigquery_column(
         self, name: str, schema_property: dict
     ) -> SchemaField:
-        """Translate a JSON schema property into a BigQuery column."""
-        if "anyOf" in schema_property and len(schema_property["anyOf"]) > 0:
-            # I have only seen this used in the wild with tap-salesforce, which
-            # is incidentally an important one so lets handle the anyOf case
-            # by giving the 0th index priority.
-            property_type = schema_property["anyOf"][0].get("type", "string")
-            property_format = schema_property["anyOf"][0].get("format", None)
-        else:
-            property_type = schema_property.get("type", "string")
-            property_format = schema_property.get("format", None)
+        """Translate a JSON schema property into a BigQuery column.
+        
+        The v1 resolver is very similar to how existing target-bigquery implementations worked.
+        The v2 resolver uses JSON in all cases the schema property is unresolvable making it _much_
+        more flexible though the denormalization can only be said to be partial if a type is not
+        resolved. Most of the time this is fine but for the sake of consistency, we default to v1.
+        """
+        if self.resolver_version == SchemaResolverVersion.V1:
+            # This is the original resolver, which is used by the denormalized strategy
+            if "anyOf" in schema_property and len(schema_property["anyOf"]) > 0:
+                # I have only seen this used in the wild with tap-salesforce, which
+                # is incidentally an important one so lets handle the anyOf case
+                # by giving the 0th index priority.
+                property_type = schema_property["anyOf"][0].get("type", "string")
+                property_format = schema_property["anyOf"][0].get("format", None)
+            else:
+                property_type = schema_property.get("type", "string")
+                property_format = schema_property.get("format", None)
+
+            if "array" in property_type:
+                if "items" not in schema_property:
+                    return SchemaField(name, "JSON", "REPEATED")
+                items_schema: dict = schema_property["items"]
+                items_type = bigquery_type(items_schema["type"], items_schema.get("format", None))
+                if items_type == "record":
+                    return self._translate_record_to_bigquery_schema(name, items_schema, "REPEATED")
+                return SchemaField(name, items_type, "REPEATED")
+            elif "object" in property_type:
+                return self._translate_record_to_bigquery_schema(name, schema_property)
+            else:
+                result_type = bigquery_type(property_type, property_format)
+                return SchemaField(name, result_type, "NULLABLE")
+        elif self.resolver_version == SchemaResolverVersion.V2:
+            # This is the new resolver, which is far more lenient and falls back to JSON
+            # if it doesn't know how to translate a property.
+            try:
+                if "anyOf" in schema_property and len(schema_property["anyOf"]) > 0:
+                    # I have only seen this used in the wild with tap-salesforce, which
+                    # is incidentally an important one so lets handle the anyOf case
+                    # by giving the 0th index priority.
+                    property_type = schema_property["anyOf"][0].get("type", "string")
+                    property_format = schema_property["anyOf"][0].get("format", None)
+                else:
+                    property_type = schema_property["type"]
+                    property_format = schema_property.get("format", None)
+
+                if "array" in property_type:
+                    if "items" not in schema_property or "type" not in schema_property["items"]:
+                        return SchemaField(name, "JSON", "REPEATED")
+                    items_schema: dict = schema_property["items"]
+                    if "patternProperties" in items_schema:
+                        return SchemaField(name, "JSON", "REPEATED")
+                    items_type = bigquery_type(items_schema["type"], items_schema.get("format", None))
+                    if items_type == "record":
+                        return self._translate_record_to_bigquery_schema(name, items_schema, "REPEATED")
+                    return SchemaField(name, items_type, "REPEATED")
+                elif "object" in property_type:
+                    if "properties" not in schema_property or len(schema_property["properties"]) == 0 or "patternProperties" in schema_property:
+                        return SchemaField(name, "JSON", "NULLABLE")
+                    return self._translate_record_to_bigquery_schema(name, schema_property)
+                else:
+                    if "patternProperties" in schema_property:
+                        return SchemaField(name, "JSON", "NULLABLE")
+                    result_type = bigquery_type(property_type, property_format)
+                    return SchemaField(name, result_type, "NULLABLE")
+            except Exception:
+                return SchemaField(name, "JSON", "NULLABLE")
 
-        if "array" in property_type:
-            if "items" not in schema_property:
-                return SchemaField(name, "JSON", "REPEATED")
-            items_schema: dict = schema_property["items"]
-            items_type = bigquery_type(items_schema["type"], items_schema.get("format", None))
-            if items_type == "record":
-                return self._translate_record_to_bigquery_schema(name, items_schema, "REPEATED")
-            return SchemaField(name, items_type, "REPEATED")
-        elif "object" in property_type:
-            return self._translate_record_to_bigquery_schema(name, schema_property)
-        else:
-            result_type = bigquery_type(property_type, property_format)
-            return SchemaField(name, result_type, "NULLABLE")
 
     def _translate_record_to_bigquery_schema(
         self, name: str, schema_property: dict, mode: str = "NULLABLE"
@@ -692,66 +795,73 @@ class SchemaTranslator:
 
     def _bigquery_field_to_projection(
         self, field: SchemaField, path: str = "$", depth: int = 0, base: str = "data"
-    ) -> str:
+    ) -> _FieldProjection:
         """Translate a BigQuery schema field into a SQL projection."""
         # Pass-through _sdc columns into the projection as-is
         if field.name.startswith("_sdc_"):
-            return f"{field.name},\n"
+            return _FieldProjection(field.name, field.name)
 
         scalar = f"{base}.{field.name}"
         from_base = f"{path}.{field.name}" if base == "data" else "$"
 
         # Records are handled recursively
         if field.field_type.upper() == "RECORD":
-            return (
+            return _FieldProjection(
                 (" " * depth * 2)
                 + "STRUCT(\n{}\n".format(
                     "".join(
                         [
-                            self._bigquery_field_to_projection(
-                                f, path=from_base, depth=depth + 1, base=base
+                            (
+                                self._bigquery_field_to_projection(
+                                    f, path=from_base, depth=depth + 1, base=base
+                                ).as_sql()
+                                if not f.mode == "REPEATED"
+                                else self._wrap_json_array(
+                                    f, path=from_base, depth=depth, base=base
+                                )
                             )
-                            if not f.mode == "REPEATED"
-                            else self._wrap_json_array(f, path=from_base, depth=depth, base=base)
                             for f in field.fields
                         ]
                     ).rstrip(",\n"),
                 )
                 + (" " * depth * 2)
-                + f") as {transform_column_name(field.name, **self.transforms)},\n"
+                + ")", f"{transform_column_name(field.name, **self.transforms)}"
             )
         # Nullable fields require a JSON_VALUE call which creates a 2-stage cast
         elif field.is_nullable:
-            return (" " * depth * 2) + self._wrap_nullable_json_value(field, path=path, base=base)
+            _field = self._wrap_nullable_json_value(field, path=path, base=base)
+            return _FieldProjection((" " * depth * 2) + _field.projection, _field.alias)
         # These are not nullable so if the type is known, we can do a 1-stage extract & cast
         elif field.field_type.upper() == "STRING":
-            return (
+            return _FieldProjection((
                 " " * depth * 2
-            ) + f"STRING({scalar}) as {transform_column_name(field.name, **self.transforms)},\n"
+            ) + f"STRING({scalar})", f"{transform_column_name(field.name, **self.transforms)}")
         elif field.field_type.upper() == "INTEGER":
-            return (
+            return _FieldProjection((
                 " " * depth * 2
-            ) + f"INT64({scalar}) as {transform_column_name(field.name, **self.transforms)},\n"
+            ) + f"INT64({scalar})", f"{transform_column_name(field.name, **self.transforms)}")
         elif field.field_type.upper() == "FLOAT":
-            return (
+            return _FieldProjection(
                 (" " * depth * 2)
-                + f"FLOAT64({scalar}) as {transform_column_name(field.name, **self.transforms)},\n"
+                + f"FLOAT64({scalar})", f"{transform_column_name(field.name, **self.transforms)}"
             )
         elif field.field_type.upper() == "BOOLEAN":
-            return (
+            return _FieldProjection((
                 " " * depth * 2
-            ) + f"BOOL({scalar}) as {transform_column_name(field.name, **self.transforms)},\n"
+            ) + f"BOOL({scalar})", f"{transform_column_name(field.name, **self.transforms)}")
         # Fallback to a 2-stage extract & cast
         else:
-            return (" " * depth * 2) + self._wrap_nullable_json_value(field, path, base)
+            _field =  self._wrap_nullable_json_value(field, path, base)
+            return _FieldProjection((" " * depth * 2) + _field.projection, _field.alias)
 
     def _wrap_json_array(
         self, field: SchemaField, path: str, depth: int = 0, base: str = "data"
     ) -> str:
         """Translate a BigQuery schema field into a SQL projection for a repeated field."""
-        v = self._bigquery_field_to_projection(
+        _v = self._bigquery_field_to_projection(
             field, path="$", depth=depth, base=f"{field.name}__rows"
-        ).rstrip(", \n")
+        )
+        v = _v.as_sql().rstrip(", \n")
         return (" " * depth * 2) + indent(
             dedent(
                 f"""
@@ -760,6 +870,7 @@ class SchemaTranslator:
             FROM UNNEST(
                 JSON_QUERY_ARRAY({base}, '{path}.{field.name}')
             ) AS {field.name}__rows
+            WHERE {_v.projection} IS NOT NULL
         """
                 + (" " * depth * 2)
                 + f") AS {field.name},\n"
@@ -769,21 +880,21 @@ class SchemaTranslator:
 
     def _wrap_nullable_json_value(
         self, field: SchemaField, path: str = "$", base: str = "data"
-    ) -> str:
+    ) -> _FieldProjection:
         """Translate a BigQuery schema field into a SQL projection for a nullable field."""
         typ = field.field_type.upper()
         if typ == "STRING":
-            return (
-                f"JSON_VALUE({base}, '{path}.{field.name}') as"
-                f" {transform_column_name(field.name, **self.transforms)},\n"
+            return _FieldProjection(
+                f"JSON_VALUE({base}, '{path}.{field.name}')",
+                f" {transform_column_name(field.name, **self.transforms)}"
             )
         if typ == "FLOAT":
             typ = "FLOAT64"
         if typ in ("INT", "INTEGER"):
             typ = "INT64"
-        return (
-            f"CAST(JSON_VALUE({base}, '{path}.{field.name}') as {typ}) as"
-            f" {transform_column_name(field.name, **self.transforms)},\n"
+        return _FieldProjection(
+            f"CAST(JSON_VALUE({base}, '{path}.{field.name}') as {typ})",
+            f" {transform_column_name(field.name, **self.transforms)}"
         )
 
 
