@@ -9,13 +9,15 @@
 # The above copyright notice and this permission notice shall be included in all copies or
 # substantial portions of the Software.
 """BigQuery target class."""
+import os
 import copy
 import time
 import uuid
 from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Type, Union
 
+from singer_sdk import Sink
 from singer_sdk import typing as th
-from singer_sdk.target_base import Sink, Target
+from singer_sdk.target_base import Target
 
 from target_bigquery.batch_job import BigQueryBatchJobDenormalizedSink, BigQueryBatchJobSink
 from target_bigquery.core import BaseBigQuerySink, BaseWorker, BigQueryCredentials, ParType
@@ -36,15 +38,18 @@ if TYPE_CHECKING:
 # Defaults for target worker pool parameters
 MAX_WORKERS = 15
 """Maximum number of workers to spawn."""
+MAX_JOBS_QUEUED = 30
+"""Maximum number of jobs placed in the global queue to avoid memory overload."""
 WORKER_CAPACITY_FACTOR = 5
 """Jobs enqueued must exceed the number of active workers times this number."""
 WORKER_CREATION_MIN_INTERVAL = 5
 """Minimum time between worker creation attempts."""
 
-
 class TargetBigQuery(Target):
     """Target for BigQuery."""
 
+    _MAX_RECORD_AGE_IN_MINUTES = 5.0
+    
     name = "target-bigquery"
     config_jsonschema = th.PropertiesList(
         th.Property(
@@ -81,6 +86,12 @@ class TargetBigQuery(Target):
             description="The maximum number of rows to send in a single batch or commit.",
             default=500,
         ),
+        th.Property(
+            "fail_fast",
+            th.BooleanType,
+            description="Fail the entire load job if any row fails to insert.",
+            default=True,
+        ),
         th.Property(
             "timeout",
             th.IntegerType,
@@ -220,8 +231,8 @@ class TargetBigQuery(Target):
                     th.IntegerType,
                     required=False,
                     description=(
-                        "By default, each sink type has a preconfigured max worker limit. This sets"
-                        " an override for maximum number of workers."
+                        "By default, each sink type has a preconfigured max worker pool limit."
+                        " This sets an override for maximum number of workers in the pool."
                     ),
                 ),
             ),
@@ -293,6 +304,16 @@ class TargetBigQuery(Target):
                 " from different extracts."
             ),
         ),
+        th.Property(
+            "schema_resolver_version",
+            th.IntegerType,
+            default=1,
+            description=(
+                "The version of the schema resolver to use. Defaults to 1. Version 2 uses JSON as a"
+                " fallback during denormalization. This only has an effect if denormalized=true"
+            ),
+            allowed_values=[1, 2],
+        ),
     ).to_dict()
 
     def __init__(self, *args, **kwargs) -> None:
@@ -306,6 +327,8 @@ class TargetBigQuery(Target):
         ) = self.get_parallelization_components()
         self.queue = self.queue_cls()
         self.job_notification, self.job_notifier = self.pipe_cls(False)
+        self.log_notification, self.log_notifier = self.pipe_cls(False)
+        self.error_notification, self.error_notifier = self.pipe_cls(False)
         self._credentials = BigQueryCredentials(
             self.config.get("credentials_path"),
             self.config.get("credentials_json"),
@@ -313,11 +336,16 @@ class TargetBigQuery(Target):
         )
 
         def worker_factory():
-            return self.get_sink_class().worker_cls_factory(self.proc_cls, self.config,)(
+            return self.get_sink_class().worker_cls_factory(
+                self.proc_cls,
+                self.config,
+            )(
                 ext_id=uuid.uuid4().hex,
                 queue=self.queue,
                 credentials=self._credentials,
                 job_notifier=self.job_notifier,
+                log_notifier=self.log_notifier,
+                error_notifier=self.error_notifier,
             )
 
         self.worker_factory = worker_factory
@@ -459,12 +487,31 @@ class TargetBigQuery(Target):
         return existing_sink
 
     def drain_one(self, sink: Sink) -> None:  # type: ignore
-        """Drain a sink. Includes a hook to manage the worker pool."""
+        """Drain a sink. Includes a hook to manage the worker pool and notifications."""
+        #self.logger.info(f"Jobs queued : {self.queue.qsize()} | Max nb jobs queued : {os.cpu_count() * 4} | Nb workers : {len(self.workers)} | Max nb workers : {os.cpu_count() * 2}")
         self.resize_worker_pool()
         while self.job_notification.poll():
             ext_id = self.job_notification.recv()
             self.worker_pings[ext_id] = time.time()
             self._jobs_enqueued -= 1
+        while self.log_notification.poll():
+            msg = self.log_notification.recv()
+            self.logger.info(msg)
+        if self.error_notification.poll():
+            e, msg = self.error_notification.recv()
+            if self.config.get("fail_fast", True):
+                self.logger.error(msg)
+                try:
+                    # Try to drain if we can. This is a best effort.
+                    # TODO: we should consider if draining here is the right thing
+                    # to do. It's _possible_ we increment the state message when
+                    # data is not actually written. Its _unlikely_ so the upside is
+                    # greater than the downside for now but will revisit this.
+                    self.logger.error("Draining all sinks and terminating.")
+                    self.drain_all(is_endofpipe=True)
+                except Exception:
+                    self.logger.error("Drain failed.")
+                raise RuntimeError(msg) from e
         super().drain_one(sink)
 
     def drain_all(self, is_endofpipe: bool = False) -> None:  # type: ignore
@@ -477,11 +524,14 @@ class TargetBigQuery(Target):
             for worker in self.workers:
                 if worker.is_alive():
                     self.queue.put(None)
-            for worker in self.workers:
+            while len(self.workers):
                 worker.join()
+                worker = self.workers.pop()
             for sink in self._sinks_active.values():
                 sink.clean_up()
         else:
+            for worker in self.workers:
+                worker.join()
             for sink in self._sinks_active.values():
                 sink.pre_state_hook()
         self._write_state_message(state)
